import pandas as pd
import os, sys, pickle, subprocess, shutil
from sklearn.model_selection import KFold
from pathlib import Path

from GANDLF.compute import training_loop
from GANDLF.utils import get_dataframe


def TrainingManager(dataframe, outputDir, parameters, device, resume, reset):
    """
    This is the training manager that ties all the training functionality together

    Args:
        dataframe (pandas.DataFrame): The full data from CSV.
        outputDir (str): The main output directory.
        parameters (dict): The parameters dictionary.
        device (str): The device to perform computations on.
        resume (bool): Whether the previous run will be resumed or not.
        reset (bool): Whether the previous run will be reset or not.
    """
    if reset:
        shutil.rmtree(outputDir)
        Path(outputDir).mkdir(parents=True, exist_ok=True)

    # save the current model configuration as a sanity check
    currentModelConfigPickle = os.path.join(outputDir, "parameters.pkl")
    if (not os.path.exists(currentModelConfigPickle)) or reset or resume:
        with open(currentModelConfigPickle, "wb") as handle:
            pickle.dump(parameters, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        if os.path.exists(currentModelConfigPickle):
            print(
                "Using previously saved parameter file",
                currentModelConfigPickle,
                flush=True,
            )
            parameters = pickle.load(open(currentModelConfigPickle, "rb"))

    # check for single fold training
    singleFoldValidation = False
    singleFoldTesting = False
    noTestingData = False
    # if the user wants a single fold training
    if parameters["nested_training"]["testing"] < 0:
        parameters["nested_training"]["testing"] = abs(
            parameters["nested_training"]["testing"]
        )
        singleFoldTesting = True

    # if the user wants a single fold training
    if parameters["nested_training"]["validation"] < 0:
        parameters["nested_training"]["validation"] = abs(
            parameters["nested_training"]["validation"]
        )
        singleFoldValidation = True

    # this is the condition where testing data is not to be kept
    if parameters["nested_training"]["testing"] == 1:
        noTestingData = True
        singleFoldTesting = True
        # put 2 just so that the first for-loop does not fail
        parameters["nested_training"]["testing"] = 2

    # initialize the kfold structures
    kf_testing = KFold(n_splits=parameters["nested_training"]["testing"])
    kf_validation = KFold(n_splits=parameters["nested_training"]["validation"])

    currentTestingFold = 0

    # split across subjects
    subjectIDs_full = (
        dataframe[dataframe.columns[parameters["headers"]["subjectIDHeader"]]]
        .unique()
        .tolist()
    )

    # get the indeces for kfold splitting
    trainingData_full = dataframe

    # start the kFold train for testing
    for trainAndVal_index, testing_index in kf_testing.split(subjectIDs_full):
        # ensure the validation fold is initialized per-testing split
        currentValidationFold = 0

        trainingAndValidationData = pd.DataFrame()  # initialize the variable
        testingData = pd.DataFrame()  # initialize the variable
        # get the current training and testing data
        if noTestingData:
            # don't consider the split indeces for this case
            trainingAndValidationData = trainingData_full
            testingData = None
        else:
            # loop over all trainAndVal_index and construct new dataframe
            for subject_idx in trainAndVal_index:
                trainingAndValidationData = trainingAndValidationData._append(
                    trainingData_full[
                        trainingData_full[
                            trainingData_full.columns[
                                parameters["headers"]["subjectIDHeader"]
                            ]
                        ]
                        == subjectIDs_full[subject_idx]
                    ]
                )

            # loop over all testing_index and construct new dataframe
            for subject_idx in testing_index:
                testingData = testingData._append(
                    trainingData_full[
                        trainingData_full[
                            trainingData_full.columns[
                                parameters["headers"]["subjectIDHeader"]
                            ]
                        ]
                        == subjectIDs_full[subject_idx]
                    ]
                )

        # the output of the current fold is only needed if multi-fold training is happening
        if singleFoldTesting:
            currentOutputFolder = outputDir
        else:
            currentOutputFolder = os.path.join(
                outputDir, "testing_" + str(currentTestingFold)
            )
            Path(currentOutputFolder).mkdir(parents=True, exist_ok=True)

        # save the current training+validation and testing datasets
        if noTestingData:
            print(
                "WARNING: Testing data is empty, which will result in scientifically incorrect results; use at your own risk."
            )
            current_training_subject_indeces_full = subjectIDs_full
            currentTestingDataPickle = "None"
        else:
            currentTrainingAndValidationDataPickle = os.path.join(
                currentOutputFolder, "data_trainAndVal.pkl"
            )
            currentTestingDataPickle = os.path.join(
                currentOutputFolder, "data_testing.pkl"
            )

            if (not os.path.exists(currentTestingDataPickle)) or reset or resume:
                testingData.to_pickle(currentTestingDataPickle)
            else:
                if os.path.exists(currentTestingDataPickle):
                    print(
                        "Using previously saved testing data",
                        currentTestingDataPickle,
                        flush=True,
                    )
                    testingData = pd.read_pickle(currentTestingDataPickle)

            if (
                (not os.path.exists(currentTrainingAndValidationDataPickle))
                or reset
                or resume
            ):
                trainingAndValidationData.to_pickle(
                    currentTrainingAndValidationDataPickle
                )
            else:
                if os.path.exists(currentTrainingAndValidationDataPickle):
                    print(
                        "Using previously saved training+validation data",
                        currentTrainingAndValidationDataPickle,
                        flush=True,
                    )
                    trainingAndValidationData = pd.read_pickle(
                        currentTrainingAndValidationDataPickle
                    )

            current_training_subject_indeces_full = (
                trainingAndValidationData[
                    trainingAndValidationData.columns[
                        parameters["headers"]["subjectIDHeader"]
                    ]
                ]
                .unique()
                .tolist()
            )

        # start the kFold train for validation
        for train_index, val_index in kf_validation.split(
            current_training_subject_indeces_full
        ):
            # the output of the current fold is only needed if multi-fold training is happening
            if singleFoldValidation:
                currentValOutputFolder = currentOutputFolder
            else:
                currentValOutputFolder = os.path.join(
                    currentOutputFolder, str(currentValidationFold)
                )
                Path(currentValOutputFolder).mkdir(parents=True, exist_ok=True)

            trainingData = pd.DataFrame()  # initialize the variable
            validationData = pd.DataFrame()  # initialize the variable

            # loop over all train_index and construct new dataframe
            for subject_idx in train_index:
                trainingData = trainingData._append(
                    trainingData_full[
                        trainingData_full[
                            trainingData_full.columns[
                                parameters["headers"]["subjectIDHeader"]
                            ]
                        ]
                        == subjectIDs_full[subject_idx]
                    ]
                )

            # loop over all val_index and construct new dataframe
            for subject_idx in val_index:
                validationData = validationData._append(
                    trainingData_full[
                        trainingData_full[
                            trainingData_full.columns[
                                parameters["headers"]["subjectIDHeader"]
                            ]
                        ]
                        == subjectIDs_full[subject_idx]
                    ]
                )

            # # write parameters to pickle - this should not change for the different folds, so keeping is independent
            ## pickle/unpickle data
            # pickle the data
            currentTrainingDataPickle = os.path.join(
                currentValOutputFolder, "data_training.pkl"
            )
            currentValidationDataPickle = os.path.join(
                currentValOutputFolder, "data_validation.pkl"
            )
            if (not os.path.exists(currentTrainingDataPickle)) or reset or resume:
                trainingData.to_pickle(currentTrainingDataPickle)
                trainingData.to_csv(
                    currentTrainingDataPickle.replace(".pkl", ".csv"), index=False
                )
            else:
                trainingData = get_dataframe(currentTrainingDataPickle)
            if (not os.path.exists(currentValidationDataPickle)) or reset or resume:
                validationData.to_pickle(currentValidationDataPickle)
                validationData.to_csv(
                    currentValidationDataPickle.replace(".pkl", ".csv"), index=False
                )
            else:
                validationData = get_dataframe(currentValidationDataPickle)

            # parallel_compute_command is an empty string, thus no parallel computing requested
            if (not parameters["parallel_compute_command"]) or (singleFoldValidation):
                training_loop(
                    training_data=trainingData,
                    validation_data=validationData,
                    output_dir=currentValOutputFolder,
                    device=device,
                    params=parameters,
                    testing_data=testingData,
                )

            else:
                # call qsub here
                parallel_compute_command_actual = parameters[
                    "parallel_compute_command"
                ].replace("${outputDir}", currentValOutputFolder)

                if not ("python" in parallel_compute_command_actual):
                    sys.exit(
                        "The 'parallel_compute_command_actual' needs to have the python from the virtual environment, which is usually '${GANDLF_dir}/venv/bin/python'"
                    )

                command = (
                    parallel_compute_command_actual
                    + " -m GANDLF.training_loop -train_loader_pickle "
                    + currentTrainingDataPickle
                    + " -val_loader_pickle "
                    + currentValidationDataPickle
                    + " -parameter_pickle "
                    + currentModelConfigPickle
                    + " -device "
                    + str(device)
                    + " -outputDir "
                    + currentValOutputFolder
                    + " -testing_loader_pickle "
                    + currentTestingDataPickle
                )

                print(
                    "Submitting job for testing split "
                    + str(currentTestingFold)
                    + " and validation split "
                    + str(currentValidationFold)
                )
                subprocess.Popen(command, shell=True).wait()

            if singleFoldValidation:
                break
            currentValidationFold += 1  # go to next fold

        if singleFoldTesting:
            break
        currentTestingFold += 1  # go to next fold


def TrainingManager_split(
    dataframe_train,
    dataframe_validation,
    dataframe_testing,
    outputDir,
    parameters,
    device,
    resume,
    reset,
):
    """
    This is the training manager that ties all the training functionality together

    Args:
        dataframe_train (pandas.DataFrame): The training data from CSV.
        dataframe_validation (pandas.DataFrame): The validation data from CSV.
        dataframe_testing (pandas.DataFrame): The testing data from CSV.
        outputDir (str): The main output directory.
        parameters (dict): The parameters dictionary.
        device (str): The device to perform computations on.
        resume (bool): Whether the previous run will be resumed or not.
        reset (bool): Whether the previous run will be reset or not.
    """
    currentModelConfigPickle = os.path.join(outputDir, "parameters.pkl")
    if (not os.path.exists(currentModelConfigPickle)) or reset or resume:
        with open(currentModelConfigPickle, "wb") as handle:
            pickle.dump(parameters, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        if os.path.exists(currentModelConfigPickle):
            print(
                "Using previously saved parameter file",
                currentModelConfigPickle,
                flush=True,
            )
            parameters = pickle.load(open(currentModelConfigPickle, "rb"))

    training_loop(
        training_data=dataframe_train,
        validation_data=dataframe_validation,
        output_dir=outputDir,
        device=device,
        params=parameters,
        testing_data=dataframe_testing,
    )



import torchmetrics as tm
from torch.nn.functional import one_hot
from ..utils import get_output_from_calculator
from GANDLF.utils.generic import determine_classification_task_type


def overall_stats(predictions, ground_truth, params):
    """
    Generates a dictionary of metrics calculated on the overall predictions and ground truths.

    Args:
        predictions (torch.Tensor): The output of the model.
        ground_truth (torch.Tensor): The ground truth labels.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        dict: A dictionary of metrics.
    """
    assert (
        params["problem_type"] == "classification"
    ), "Only classification is supported for these stats"

    output_metrics = {}

    average_types_keys = {
        "global": "micro",
        "per_class": "none",
        "per_class_average": "macro",
        "per_class_weighted": "weighted",
    }
    task = determine_classification_task_type(params)
    # consider adding a "multilabel field in the future"
    # metrics that need the "average" parameter

    for average_type_key in average_types_keys.values():
        # multidim_average is not used when constructing these metrics
        # think of having it
        calculators = {
            "accuracy": tm.Accuracy(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key,
            ),
            "precision": tm.Precision(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key,
            ),
            "recall": tm.Recall(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key,
            ),
            "f1": tm.F1Score(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key,
            ),
            "specificity": tm.Specificity(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key,
            ),
            "aucroc": tm.AUROC(
                task=task,
                num_classes=params["model"]["num_classes"],
                average=average_type_key if average_type_key != "micro" else "macro",
            ),
        }
        for metric_name, calculator in calculators.items():
            if metric_name == "aucroc":
                one_hot_preds = one_hot(
                    predictions.long(),
                    num_classes=params["model"]["num_classes"],
                )
                output_metrics[metric_name] = get_output_from_calculator(
                    one_hot_preds.float(), ground_truth, calculator
                )
            else:
                output_metrics[metric_name] = get_output_from_calculator(
                    predictions, ground_truth, calculator
                )

    #### HERE WE NEED TO MODIFY TESTS - ROC IS RETURNING A TUPLE. WE MAY ALSO DISCRAD IT ####
    # what is AUC metric telling at all? Computing it for predictions and ground truth
    # is not making sense
    # metrics that do not have any "average" parameter
    # calculators = {
    #
    #     # "auc": tm.AUC(reorder=True),
    #     ## weird error for multi-class problem, where pos_label is not getting set
    #     "roc": tm.ROC(task=task, num_classes=params["model"]["num_classes"]),
    # }
    # for metric_name, calculator in calculators.items():
    #     if metric_name == "roc":
    #         one_hot_preds = one_hot(
    #             predictions.long(), num_classes=params["model"]["num_classes"]
    #         )
    #         output_metrics[metric_name] = get_output_from_calculator(
    #             one_hot_preds.float(), ground_truth, calculator
    #         )
    #     else:
    #         output_metrics[metric_name] = get_output_from_calculator(
    #             predictions, ground_truth, calculator
    #         )

    return output_metrics



# -*- coding: utf-8 -*-
"""All Models in GANDLF are to be derived from this base class code."""

import torch.nn as nn

from acsconv.converters import ACSConverter, Conv3dConverter, SoftACSConverter

from GANDLF.utils import get_linear_interpolation_mode
from GANDLF.utils.generic import checkPatchDimensions
from GANDLF.utils.modelbase import get_modelbase_final_layer
from GANDLF.models.seg_modules.average_pool import (
    GlobalAveragePooling3D,
    GlobalAveragePooling2D,
)


class ModelBase(nn.Module):
    """
    This is the base model class that all other architectures will need to derive from
    """

    def __init__(self, parameters):
        """
        This defines all defaults that the model base uses

        Args:
            parameters (dict): This is a dictionary of all parameters that are needed for the model.
        """
        super(ModelBase, self).__init__()
        self.model_name = parameters["model"]["architecture"]
        self.n_dimensions = parameters["model"]["dimension"]
        self.n_channels = parameters["model"]["num_channels"]
        if "num_classes" in parameters["model"]:
            self.n_classes = parameters["model"]["num_classes"]
        else:
            self.n_classes = len(parameters["model"]["class_list"])
        self.base_filters = parameters["model"]["base_filters"]
        self.norm_type = parameters["model"]["norm_type"]
        self.patch_size = parameters["patch_size"]
        self.batch_size = parameters["batch_size"]
        self.amp = parameters["model"]["amp"]
        self.final_convolution_layer = self.get_final_layer(
            parameters["model"]["final_layer"]
        )

        self.linear_interpolation_mode = get_linear_interpolation_mode(
            self.n_dimensions
        )

        self.sigmoid_input_multiplier = parameters["model"].get(
            "sigmoid_input_multiplier", 1.0
        )

        # based on dimensionality, the following need to defined:
        # convolution, batch_norm, instancenorm, dropout
        if self.n_dimensions == 2:
            self.Conv = nn.Conv2d
            self.ConvTranspose = nn.ConvTranspose2d
            self.InstanceNorm = nn.InstanceNorm2d
            self.Dropout = nn.Dropout2d
            self.BatchNorm = nn.BatchNorm2d
            self.MaxPool = nn.MaxPool2d
            self.AvgPool = nn.AvgPool2d
            self.AdaptiveAvgPool = nn.AdaptiveAvgPool2d
            self.AdaptiveMaxPool = nn.AdaptiveMaxPool2d
            self.GlobalAvgPool = GlobalAveragePooling2D
            self.Norm = self.get_norm_type(self.norm_type.lower(), self.n_dimensions)
            self.converter = None

        elif self.n_dimensions == 3:
            self.Conv = nn.Conv3d
            self.ConvTranspose = nn.ConvTranspose3d
            self.InstanceNorm = nn.InstanceNorm3d
            self.Dropout = nn.Dropout3d
            self.BatchNorm = nn.BatchNorm3d
            self.MaxPool = nn.MaxPool3d
            self.AvgPool = nn.AvgPool3d
            self.AdaptiveAvgPool = nn.AdaptiveAvgPool3d
            self.AdaptiveMaxPool = nn.AdaptiveMaxPool3d
            self.GlobalAvgPool = GlobalAveragePooling3D
            self.Norm = self.get_norm_type(self.norm_type.lower(), self.n_dimensions)

            # define 2d to 3d model converters
            converter_type = parameters["model"].get("converter_type", "soft").lower()
            self.converter = SoftACSConverter
            if converter_type == "acs":
                self.converter = ACSConverter
            elif converter_type == "conv3d":
                self.converter = Conv3dConverter

        else:
            raise ValueError(
                "GaNDLF only supports 2D and 3D computations. {}D computations are not currently supported".format(
                    self.n_dimensions
                )
            )

    def get_final_layer(self, final_convolution_layer):
        return get_modelbase_final_layer(final_convolution_layer)

    def get_norm_type(self, norm_type, dimensions):
        """
        This function gets the normalization type for the model.

        Args:
            norm_type (str): Normalization type as a string.
            dimensions (str): The dimensionality of the model.

        Returns:
            _InstanceNorm or _BatchNorm: The normalization type for the model.
        """
        if dimensions == 3:
            if norm_type == "batch":
                norm_type = nn.BatchNorm3d
            elif norm_type == "instance":
                norm_type = nn.InstanceNorm3d
            else:
                norm_type = None
        elif dimensions == 2:
            if norm_type == "batch":
                norm_type = nn.BatchNorm2d
            elif norm_type == "instance":
                norm_type = nn.InstanceNorm2d
            else:
                norm_type = None

        return norm_type

    def model_depth_check(self, parameters):
        """
        This function checks if the patch size is large enough for the model.

        Args:
            parameters (dict): The entire set of parameters for the model.

        Returns:
            int: The model depth to use.

        Raises:
            AssertionError: If the patch size is not large enough for the model.
        """
        model_depth = checkPatchDimensions(
            parameters["patch_size"], numlay=parameters["model"]["depth"]
        )

        common_msg = "The patch size is not large enough for desired depth. It is expected that each dimension of the patch size is divisible by 2^i, where i is in a integer greater than or equal to 2."
        assert model_depth >= 2, common_msg

        if model_depth != parameters["model"]["depth"] and model_depth >= 2:
            print(common_msg + " Only the first %d layers will run." % model_depth)

        return model_depth
